#####Sender and Receiver Latent Class Models.
#####LCERGM is a wrapper for the multiple start value set up
#####ergm.EM is the actual EM algorithm
library(statnet)
library(ergm.userterms)



LCERGM <- function(fixedeffects, latentclasseffects, network, k, type = "sender", n.random.starts = 20, max.iter = 100){
  nodes <- dim(as.matrix(network))[[1]]
  no.fixed <- .~.
  fit.models <- list()
  AICs <- vector()
  for(i in 1:n.random.starts){
    print(paste("Run:", i))
    start.vals <- sample(1:k, nodes, replace = T)
    inital.fit <-invisible( ergm.EM(no.fixed, latentclasseffects, network, start.vals, max.iter, k, type, T ))
    start.vals <- inital.fit[[2]]
    fit.models[[i]] <- ergm.EM(fixedeffects, latentclasseffects, network, start.vals, max.iter, k, type, F )
    AICs[i] <- extractAIC(fit.models[[i]][[1]])[2]
  }
  
  return(fit.models[[which(AICs == min(AICs))[1]]])
  
}
ergm.EM <- function(fixedeffects,latentclasseffects, network, start.vals, max.iter, k, type = "sender", initial = F){
  
  nodes <- dim(as.matrix(network))[[1]]
  class.labs <- start.vals
  class.prob <- matrix(0, ncol = k, nrow = nodes)
  
  alpha.Z <- table(class.labs)/nodes
  
  lat.dat <- ergmMPLE(latentclasseffects)
  
  n.latent.var.cols <- dim(lat.dat[[2]])[[2]]
  
  form2 <- update(latentclasseffects,network~. )
  form2 <- update(latentclasseffects, fixedeffects)
  
  if(type == "sender"){
    form2 <- update(form2, .~.+sender(base = 0))
  }else{if(type == "receiver"){
    form2 <- update(form2, .~.+receiver(base = 0))
    
    
  }else{
    stop("Type is not sender or receiver")
    
  }
  }
  dat <- ergmMPLE(form2)
  
  latent.dat <- as.data.frame(dat[[2]][,1:n.latent.var.cols])
  names(latent.dat) <- colnames(dat[[2]])[1:n.latent.var.cols]
  if(!initial){ fixed.dat <- as.data.frame(dat[[2]][,(n.latent.var.cols+1):(dim(dat[[2]])[[2]] - nodes)])
  names(fixed.dat) <- colnames(dat[[2]])[(n.latent.var.cols+1):(dim(dat[[2]])[[2]] - nodes)]
  }
  
  
  
  
  
  response <- dat[[1]]
  weights<- dat[[3]]
  if(!initial){
    data.for.analysis <-data.frame(y = response, lat = I(as.matrix(latent.dat)), fix = I(as.matrix(fixed.dat)))
  }else{
    data.for.analysis <-data.frame(y = response, lat = I(as.matrix(latent.dat)))
    
  }
  lengthofpred <- dim(dat[[2]])[[2]]
  lengthoffixed <- lengthofpred - nodes - n.latent.var.cols
  expanded.labs <- vector()
  length(expanded.labs) <- dim(latent.dat)[[1]]
  index.mat <- matrix(F, nrow = dim(dat[[2]])[[1]], ncol = nodes)
  
  ll.curr <-NA
  
  
  for(i in 1:nodes){
    index.mat[,i] <- (dat[[2]][,lengthofpred-nodes+i] == 1)
  }
  
  toggle = T
  iter = 0
  while(toggle){
    
    
    for(i in 1:nodes){
      expanded.labs[index.mat[,i]] <- class.labs[i]
    }
    
    data.for.analysis[["LCLabs"]] <- I(as.factor(expanded.labs))
    
    if(!initial){
      model.fit <- glm(y ~ lat:LCLabs + fix -1, data = data.for.analysis, family = "binomial", weights = weights)
    }else{
      model.fit <- glm(y ~ lat:LCLabs - 1, data = data.for.analysis, family = "binomial", weights = weights)      
    }
    ll.now <- logLik(model.fit)
    coefs <- model.fit$coef
    
    for(i in 1:nodes){
      lls <- vector()
      node.response.dat <- subset(response, index.mat[,i])
      node.weight.dat <- subset(weights, index.mat[,i])
      node.latent.dat <- subset(latent.dat, index.mat[,i])
      if(!initial){
        node.fixed.dat <- subset(fixed.dat, index.mat[,i])
      }
      
      lab.prob <- vector()
      for(q in 1:k){
        if(q == 1){
          latent.pred.mat <-  node.latent.dat
        }else{
          latent.pred.mat <-  matrix(0,dim(node.latent.dat)[[1]], n.latent.var.cols)
        }
        for(l in 2:k){
          if(l == q){
            latent.pred.mat <-  cbind(latent.pred.mat,node.latent.dat)
          }else{
            latent.pred.mat <-  cbind(latent.pred.mat,matrix(0,dim(node.latent.dat)[[1]], n.latent.var.cols))
          }
          
        }
        if(!initial){
          node.pred <- cbind(node.fixed.dat, latent.pred.mat)
        }else{
          node.pred <- cbind(latent.pred.mat)
        }
        odds <- sweep(node.pred,MARGIN = 2,coefs,FUN = "*")
        odds <- rowSums(odds)
        normconst <- log(1+exp(odds))
        odds <- odds*node.weight.dat
        normconst <- normconst*node.weight.dat
        sum1 <- sum(odds[which(node.response.dat== 1)])
        sum2 <- sum(normconst)
        
        sum3 <- log(alpha.Z[k])
        
        lab.prob[q] <- sum1+sum3-sum2
      }
      lls <-exp(lab.prob-log(sum(exp(lab.prob))))
      
      class.prob[i,] <-lls 
      class.labs[i] <-which(lls ==max(lls))
      
      
    }
    missing.toggle <- F
    missing <- !(1:k %in%  as.numeric(names(table(class.labs))))
    missing <- which(missing == T)
    if(length(missing) > 0){
      missing.toggle = T
      toggle.missing <-missing
      for(p in 1:length(toggle.missing)){
        class.labs[sample(1:nodes, floor(nodes*.05))] <- toggle.missing[p]
      }
    }
    
    if(!is.na(ll.curr)){
      
      change.in.ll <- ll.curr-ll.now
      prop.change <- change.in.ll/ll.curr
      
      if((prop.change == 0) & !missing.toggle){
        toggle <- F
      }
      
    }
    ll.curr <- ll.now
    iter= iter+1
    if(iter >= max.iter){
      print("Maximum Iterations Achieved: Fit Suspect")
      toggle = F
    }
    
  }
  
  return(list(model = model.fit, c.labs = class.labs, c.prob = class.prob))
  
  
  
}