/*  File src/changestats.users.c in package ergm.userterms, part of the Statnet suite
 *  of packages for network analysis, https://statnet.org .
 *
 *  This software is distributed under the GPL-3 license.  It is free,
 *  open source, and has the attribution requirements (GPL Section 7) at
 *  https://statnet.org/attribution
 *
 *  Copyright 2012-2019 Statnet Commons
 */
#include "changestats.users.h"

CHANGESTAT_FN(d_mindegree) {
  Vertex t, h, node3;
  int i, mindeg, hdeg, tdeg;
  Edge e;
  int attrflag;
  double t_nodecov, h_nodecov;

  ZERO_ALL_CHANGESTATS(i);
  FOR_EACH_TOGGLE(i) {
    t = TAIL(i); h = HEAD(i);
    attrflag = INPUT_PARAM[0];
    mindeg = INPUT_PARAM[1];
    if(attrflag==0){
      tdeg = IN_DEG[t]+OUT_DEG[t];
      hdeg = IN_DEG[h]+OUT_DEG[h];
      CHANGE_STAT[0] += IS_UNDIRECTED_EDGE(t,h) ?
        - (tdeg==mindeg) - (hdeg==mindeg) :
        (tdeg==mindeg-1) + (hdeg==mindeg-1);
    }else{
      t_nodecov = INPUT_PARAM[t+1];
      h_nodecov = INPUT_PARAM[h+1];
      if (t_nodecov == h_nodecov) {
        tdeg = 0;
        STEP_THROUGH_OUTEDGES(t, e, node3) { /* step through outedges of tail */
          if(INPUT_PARAM[node3+1]==t_nodecov){++tdeg;}
        }
        STEP_THROUGH_INEDGES(t, e, node3) { /* step through inedges of tail */
          if(INPUT_PARAM[node3+1]==t_nodecov){++tdeg;}
        }
        hdeg = 0;
        STEP_THROUGH_OUTEDGES(h, e, node3) { /* step through outedges of head */
          if(INPUT_PARAM[node3+1]==h_nodecov){++hdeg;}
        }
        STEP_THROUGH_INEDGES(h, e, node3) { /* step through inedges of head */
          if(INPUT_PARAM[node3+1]==h_nodecov){++hdeg;}
        }
        CHANGE_STAT[0] += IS_UNDIRECTED_EDGE(t,h) ?
          - (tdeg==mindeg) - (hdeg==mindeg) :
          (tdeg==mindeg-1) + (hdeg==mindeg-1);
      }else{
        CHANGE_STAT[0] = 0;
      }
    }
    TOGGLE_IF_MORE_TO_COME(i);
  }
  UNDO_PREVIOUS_TOGGLES(i);
}

D_CHANGESTAT_FN(d_mixnodeicov) { 
  double sum;
  Vertex tail, head;
  int i, edgeflag, nnodes, currnode;
  
  
  /* *** don't forget tail -> head */    
  ZERO_ALL_CHANGESTATS(i);
  FOR_EACH_TOGGLE(i) 
    { 
	  nnodes = INPUT_ATTRIB[0];
      edgeflag=IS_OUTEDGE(tail = TAIL(i), head = HEAD(i));
      sum = INPUT_ATTRIB[head];
	  currnode = INPUT_ATTRIB[head + nnodes];
      CHANGE_STAT[currnode-1] += edgeflag ? -sum : sum;
      TOGGLE_IF_MORE_TO_COME(i);
    }
  UNDO_PREVIOUS_TOGGLES(i);
}

D_CHANGESTAT_FN(d_mixnodeocov) { 
  double sum;
  Vertex tail, head;
  int i, edgeflag, nnodes, currnode;
  
  
  
  /* *** don't forget tail -> head */    
  ZERO_ALL_CHANGESTATS(i);
  FOR_EACH_TOGGLE(i) 
    {
	  nnodes = INPUT_ATTRIB[0];
      edgeflag=IS_OUTEDGE(tail = TAIL(i), head = HEAD(i));
      sum = INPUT_ATTRIB[tail];
	  currnode = INPUT_ATTRIB[tail + nnodes];
      CHANGE_STAT[currnode-1] += edgeflag ? -sum : sum;
      TOGGLE_IF_MORE_TO_COME(i);
    }
  UNDO_PREVIOUS_TOGGLES(i);
}

D_CHANGESTAT_FN(d_mixabsdiff) { 
  double change, p;
  Vertex tail, head;
  int i, nnodes, currnode;

  /* *** don't forget tail -> head */
  ZERO_ALL_CHANGESTATS(i);
  FOR_EACH_TOGGLE(i) {
  nnodes = INPUT_ATTRIB[0];
    tail = TAIL(i); 
    head = HEAD(i);
    p = INPUT_ATTRIB[1];
    if(p==1.0){
      change = fabs(INPUT_ATTRIB[tail+2] - INPUT_ATTRIB[head+2]);
    } else {
      change = pow(fabs(INPUT_ATTRIB[tail+2] - INPUT_ATTRIB[head+2]), p);
    }
	if(INPUT_ATTRIB[2] == 1){
	currnode = INPUT_ATTRIB[tail + nnodes + 2];
	}
	if(INPUT_ATTRIB[2] == 2){
	currnode = INPUT_ATTRIB[head + nnodes + 2];
	}
	
    CHANGE_STAT[currnode - 1] += IS_OUTEDGE(tail,head) ? -change : change;
    TOGGLE_IF_MORE_TO_COME(i); /* Needed in case of multiple toggles */
  }
  UNDO_PREVIOUS_TOGGLES(i); /* Needed on exit in case of multiple toggles */
}


D_CHANGESTAT_FN(d_mixgwesp) { 
  Edge e, f;
  int i, echange, ochange;
  int L2th, L2tu, L2uh, currnode;
  Vertex tail, head, u, v;
  double alpha, oneexpa, cumchange, type;
  
  ZERO_ALL_CHANGESTATS(i);
  alpha = INPUT_PARAM[0];
  type = INPUT_PARAM[1];
  oneexpa = 1.0-exp(-alpha);
  
  /* *** don't forget tail -> head */    
  FOR_EACH_TOGGLE(i){      
    cumchange=0.0;
    L2th=0;
    ochange = IS_OUTEDGE(tail=TAIL(i), head=HEAD(i)) ? -1 : 0;
    echange = 2*ochange + 1;
    /* step through outedges of head  */
    STEP_THROUGH_OUTEDGES(head, e, u){
      if (IS_OUTEDGE(tail, u)){
        L2tu=ochange;
        /* step through inedges of u */
        STEP_THROUGH_INEDGES(u, f, v){
          if(IS_OUTEDGE(tail, v)) L2tu++;
        }
        cumchange += pow(oneexpa,(double)L2tu);
      }
    }
    /* step through inedges of head */
    
    STEP_THROUGH_INEDGES(head, e, u){
      if (IS_OUTEDGE(tail, u)){
        L2th++;
      }
      if (IS_OUTEDGE(u, tail)){
        L2uh=ochange;
        /* step through outedges of u */
        STEP_THROUGH_OUTEDGES(u, f, v){
          if(IS_OUTEDGE(v, head)) L2uh++;
        }
        cumchange += pow(oneexpa,(double)L2uh) ;
      }
    }
    
    if(alpha < 100.0){
      cumchange += exp(alpha)*(1.0-pow(oneexpa,(double)L2th)) ;
    }else{
      cumchange += (double)L2th;
    }
    cumchange  = echange*cumchange;
    
    if(type == 1){
      currnode = INPUT_ATTRIB[tail + 1];
    }
    if(type == 2){
      currnode = INPUT_ATTRIB[head + 1];
    }
    
    
    
    CHANGE_STAT[currnode - 1] += cumchange;
    TOGGLE_IF_MORE_TO_COME(i);
  }
  UNDO_PREVIOUS_TOGGLES(i);
}
